
#ifndef KSEEXPR_EXPORT_H
#define KSEEXPR_EXPORT_H

#ifdef KSEEXPR_STATIC_DEFINE
#  define KSEEXPR_EXPORT
#  define KSEEXPR_NO_EXPORT
#else
#  ifndef KSEEXPR_EXPORT
#    ifdef KSeExpr_EXPORTS
        /* We are building this library */
#      define KSEEXPR_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define KSEEXPR_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef KSEEXPR_NO_EXPORT
#    define KSEEXPR_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef KSEEXPR_DEPRECATED
#  define KSEEXPR_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef KSEEXPR_DEPRECATED_EXPORT
#  define KSEEXPR_DEPRECATED_EXPORT KSEEXPR_EXPORT KSEEXPR_DEPRECATED
#endif

#ifndef KSEEXPR_DEPRECATED_NO_EXPORT
#  define KSEEXPR_DEPRECATED_NO_EXPORT KSEEXPR_NO_EXPORT KSEEXPR_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef KSEEXPR_NO_DEPRECATED
#    define KSEEXPR_NO_DEPRECATED
#  endif
#endif

#endif /* KSEEXPR_EXPORT_H */
